def ONet(weight_path="onet.h5"):
    from tensorflow.keras.layers import (
        Conv2D,
        Dense,
        Flatten,
        Input,
        MaxPool2D,
        Permute,
        PReLU,
    )
    from tensorflow.keras.models import Model

    input = Input(shape=[48, 48, 3])
    x = Conv2D(32, (3, 3), strides=1, padding="valid", name="conv1")(input)
    x = PReLU(shared_axes=[1, 2], name="prelu1")(x)
    x = MaxPool2D(pool_size=3, strides=2, padding="same")(x)
    x = Conv2D(64, (3, 3), strides=1, padding="valid", name="conv2")(x)
    x = PReLU(shared_axes=[1, 2], name="prelu2")(x)
    x = MaxPool2D(pool_size=3, strides=2)(x)
    x = Conv2D(64, (3, 3), strides=1, padding="valid", name="conv3")(x)
    x = PReLU(shared_axes=[1, 2], name="prelu3")(x)
    x = MaxPool2D(pool_size=2)(x)
    x = Conv2D(128, (2, 2), strides=1, padding="valid", name="conv4")(x)
    x = PReLU(shared_axes=[1, 2], name="prelu4")(x)
    x = Permute((3, 2, 1))(x)
    x = Flatten()(x)
    x = Dense(256, name="conv5")(x)
    x = PReLU(name="prelu5")(x)

    classifier = Dense(2, activation="softmax", name="conv6-1")(x)
    bbox_regress = Dense(4, name="conv6-2")(x)
    landmark_regress = Dense(10, name="conv6-3")(x)
    model = Model([input], [classifier, bbox_regress, landmark_regress])
    model.load_weights(weight_path, by_name=True)

    return model


def RNet(weight_path="rnet.h5"):
    from tensorflow.keras.layers import (
        Conv2D,
        Dense,
        Flatten,
        Input,
        MaxPool2D,
        Permute,
        PReLU,
    )
    from tensorflow.keras.models import Model

    input = Input(
        shape=[24, 24, 3]
    )  # change this shape to [None,None,3] to enable arbitraty shape input
    x = Conv2D(28, (3, 3), strides=1, padding="valid", name="conv1")(input)
    x = PReLU(shared_axes=[1, 2], name="prelu1")(x)
    x = MaxPool2D(pool_size=3, strides=2, padding="same")(x)

    x = Conv2D(48, (3, 3), strides=1, padding="valid", name="conv2")(x)
    x = PReLU(shared_axes=[1, 2], name="prelu2")(x)
    x = MaxPool2D(pool_size=3, strides=2)(x)

    x = Conv2D(64, (2, 2), strides=1, padding="valid", name="conv3")(x)
    x = PReLU(shared_axes=[1, 2], name="prelu3")(x)
    x = Permute((3, 2, 1))(x)
    x = Flatten()(x)
    x = Dense(128, name="conv4")(x)
    x = PReLU(name="prelu4")(x)
    classifier = Dense(2, activation="softmax", name="conv5-1")(x)
    bbox_regress = Dense(4, name="conv5-2")(x)
    model = Model([input], [classifier, bbox_regress])
    model.load_weights(weight_path, by_name=True)
    return model


def PNet(weight_path="pnet.h5"):
    from tensorflow.keras.layers import (
        Conv2D,
        Dense,
        Flatten,
        Input,
        MaxPool2D,
        Permute,
        PReLU,
    )
    from tensorflow.keras.models import Model

    input = Input(shape=[None, None, 3])
    x = Conv2D(10, (3, 3), strides=1, padding="valid", name="conv1")(input)
    x = PReLU(shared_axes=[1, 2], name="PReLU1")(x)
    x = MaxPool2D(pool_size=2)(x)
    x = Conv2D(16, (3, 3), strides=1, padding="valid", name="conv2")(x)
    x = PReLU(shared_axes=[1, 2], name="PReLU2")(x)
    x = Conv2D(32, (3, 3), strides=1, padding="valid", name="conv3")(x)
    x = PReLU(shared_axes=[1, 2], name="PReLU3")(x)
    classifier = Conv2D(2, (1, 1), activation="softmax", name="conv4-1")(x)
    bbox_regress = Conv2D(4, (1, 1), name="conv4-2")(x)
    model = Model([input], [classifier, bbox_regress])
    model.load_weights(weight_path, by_name=True)
    return model



import tensorflow as tf

from .op import conv_bn, conv_dw, separable_conv
from .utils import post_processing

conf_threshold = 0.6
nms_iou_threshold = 0.3
nms_max_output_size = 200
top_k = 100
center_variance = 0.1
size_variance = 0.2

image_size = [320, 240]  # default input size 320*240
feature_map_wh_list = [[40, 30], [20, 15], [
    10, 8], [5, 4]]  # default feature map size
min_boxes = [[10, 16, 24], [32, 48], [64, 96], [128, 192, 256]]


def create_slim_net(input_shape=(240, 320), base_channel=8 * 2, num_classes=2):
    input_node = tf.keras.layers.Input(
        shape=(input_shape[0], input_shape[1], 3))

    net = conv_bn(input_node, base_channel, stride=2,
                  prefix="basenet.0")  # 120x160
    net = conv_dw(net, base_channel * 2, stride=1, prefix="basenet.1")
    net = conv_dw(net, base_channel * 2, stride=2, prefix="basenet.2")  # 60x80
    net = conv_dw(net, base_channel * 2, stride=1, prefix="basenet.3")
    net = conv_dw(net, base_channel * 4, stride=2, prefix="basenet.4")  # 30x40
    net = conv_dw(net, base_channel * 4, stride=1, prefix="basenet.5")
    net = conv_dw(net, base_channel * 4, stride=1, prefix="basenet.6")
    header_0 = conv_dw(net, base_channel * 4, stride=1, prefix="basenet.7")
    net = conv_dw(header_0, base_channel * 8, stride=2,
                  prefix="basenet.8")  # 15x20
    net = conv_dw(net, base_channel * 8, stride=1, prefix="basenet.9")
    header_1 = conv_dw(net, base_channel * 8, stride=1, prefix="basenet.10")
    net = conv_dw(header_1, base_channel * 16,
                  stride=2, prefix="basenet.11")  # 8x10
    header_2 = conv_dw(net, base_channel * 16, stride=1, prefix="basenet.12")

    out = tf.keras.layers.Conv2D(
        base_channel * 4, 1, padding="SAME", name="extras_convbias"
    )(header_2)
    out = tf.keras.layers.ReLU(name="extras_relu1")(out)
    out = separable_conv(
        out, base_channel * 16, kernel_size=3, stride=2, padding=1, prefix="extras_sep"
    )
    header_3 = tf.keras.layers.ReLU(name="extras_relu2")(out)

    reg_0 = separable_conv(
        header_0, 3 * 4, kernel_size=3, stride=1, padding=1, prefix="reg_0_sep"
    )
    cls_0 = separable_conv(
        header_0,
        3 * num_classes,
        kernel_size=3,
        stride=1,
        padding=1,
        prefix="cls_0_sep",
    )

    reg_1 = separable_conv(
        header_1, 2 * 4, kernel_size=3, stride=1, padding=1, prefix="reg_1_sep"
    )
    cls_1 = separable_conv(
        header_1,
        2 * num_classes,
        kernel_size=3,
        stride=1,
        padding=1,
        prefix="cls_1_sep",
    )

    reg_2 = separable_conv(
        header_2, 2 * 4, kernel_size=3, stride=1, padding=1, prefix="reg_2_sep"
    )
    cls_2 = separable_conv(
        header_2,
        2 * num_classes,
        kernel_size=3,
        stride=1,
        padding=1,
        prefix="cls_2_sep",
    )

    reg_3 = tf.keras.layers.Conv2D(
        3 * 4, kernel_size=3, padding="SAME", name="reg_3_convbias"
    )(header_3)
    cls_3 = tf.keras.layers.Conv2D(
        3 * num_classes, kernel_size=3, padding="SAME", name="cls_3_convbias"
    )(header_3)

    result = post_processing(
        [reg_0, reg_1, reg_2, reg_3],
        [cls_0, cls_1, cls_2, cls_3],
        num_classes,
        image_size,
        feature_map_wh_list,
        min_boxes,
        center_variance,
        size_variance,
    )

    model = tf.keras.Model(inputs=[input_node], outputs=[result])
    return model



from copy import deepcopy
from typing import Callable, List, Tuple, Union

from deep_utils.nlp.ner.base_operations import NEROperation


class SubTokenReplacement(NEROperation):
    def __init__(
        self,
        tokenizer: Callable[[str], List[str]] = str.split,
        sub_token="\u200c",
        replacement=" ",
        ner_type="BIO",
        others_label="o",
        make_copy=True,
    ):
        """
        This class replaces a sub-token with space and fixes the labels. The default sub-token is zwnj
        :param tokenizer:
        :param ner_type:
        :param make_copy: whether to create copy of the input tokens and labels to prevent data modification, the default is True
        :param sub_token: default is zwnj
        :param replacement: replacement
        :param others_label: default is o
        >>> sub_token_replacement = SubTokenReplacement()
        >>> tokens = ["تهران",
        ... r"نعمت\u200cآباد",
        ... "نور"]
        >>> labels = ["B-province", "B-section", "B-street"]
        >>> tokens, labels = sub_token_replacement(tokens, labels)
        >>> tokens
        ['تهران', 'نعمت', 'آباد', 'نور']
        >>> labels
        ['B-province', 'B-section', 'I-section', 'B-street']
        """
        super(SubTokenReplacement, self).__init__(ner_type=ner_type)
        self.sub_token = sub_token
        self.tokenizer = tokenizer
        self.make_copy = make_copy
        self.replacement = replacement
        self.others_label = others_label

    def __call__(
        self,
        tokens_x: Union[Tuple[str], List[str]],
        labels_y: Union[Tuple[str], List[str]],
    ):
        if self.make_copy:
            tokens_x, labels_y = deepcopy(tokens_x), deepcopy(labels_y)
        super().__call__(tokens_x, labels_y)
        token_index = 0
        while token_index < len(tokens_x):
            token = tokens_x[token_index]
            label = labels_y[token_index]
            if self.sub_token in token:
                replacement = token.replace(self.sub_token, self.replacement)
                sub_tokens = self.tokenizer(replacement)
                del tokens_x[token_index]
                del labels_y[token_index]
                if label.startswith("B-"):
                    sub_labels = [label] + [
                        "I-" + label.replace("B-", "")
                        for _ in range(len(sub_tokens) - 1)
                    ]
                elif label.startswith("I-") or label == self.others_label:
                    sub_labels = [label for _ in sub_tokens]
                else:
                    raise ValueError()

                for i, (sub_token, sub_label) in enumerate(zip(sub_tokens, sub_labels)):
                    tokens_x.insert(token_index + i, sub_token)
                    labels_y.insert(token_index + i, sub_label)

                token_index += len(sub_tokens)
                assert len(tokens_x) == len(labels_y)
                continue
            token_index += 1
        return tokens_x, labels_y



from typing import TypeVar

InstanceType = TypeVar("InstanceType")


class SQLAlchemyInserts:

    @staticmethod
    def insert(session, instance: InstanceType) -> InstanceType:
        try:
            session.add(instance)
            session.commit()
        except:
            session.rollback()
        return instance



from typing import Union
from pathlib import Path


def load_json(json_path: Union[str, Path], encoding="utf-8") -> dict:
    """
    loads a json file
    :param json_path: Path to json file to load
    :param encoding: encoding format
    :return: returns a json file
    """
    import json

    with open(json_path, encoding=encoding) as f:
        json_data = json.load(f)
    return json_data


def dump_json(
        json_path: Union[str, Path], json_object: Union[list, dict], encoding="utf-8", ensure_ascii=True
) -> None:
    """
    dumps a json file
    Args:
        json_object:
        json_path: path to json file
        encoding: encoding format
        ensure_ascii: set to False for persian characters

    Returns:

    """
    import json

    with open(json_path, mode="w", encoding=encoding) as f:
        json.dump(json_object, f, ensure_ascii=ensure_ascii)



from deep_utils.main_abs.main_config import MainConfig


class Config(MainConfig):
    def __init__(self):
        self.device = "cpu"
        self.confidence = 0.9
        self._model_name = "slim"
        self.model = None
        self.model_urls = {
            "rbf": "https://github.com/pooya-mohammadi/deep_utils/releases/download/0.2.0/version-RFB-320.pth",
            "slim": "https://github.com/pooya-mohammadi/deep_utils/releases/download/0.2.0/version-slim-320.pth"
        }
        self.model_caches = {
            "slim": "weights/vision/face_detection/ultra-light/torch/slim/version-slim-320.pth",
            "rbf": "weights/vision/face_detection/ultra-light/torch/RBF/version-RFB-320.pth"
        }
        self.download_variables = ("model",)

    @property
    def model_url(self):
        return self.model_urls[self.model_name]

    @property
    def model_cache(self):
        return self.model_caches[self.model_name]

    @property
    def model_name(self):
        return self._model_name.lower()

    @model_name.setter
    def model_name(self, model_name):
        self._model_name = model_name



import torch.nn as nn
from torchvision import models

from deep_utils.blocks.torch.blocks_torch import BlocksTorch
from deep_utils.utils.logging_utils.logging_utils import value_error_log


class TorchVisionModel(nn.Module):
    """
    A model that loads torch-vision models with an extra convolution layer at the end for classification
    """

    def __init__(
        self,
        model_name,
        num_classes,
        last_layer_nodes,
        use_pretrained=True,
        feature_extract=True,
        logger=None,
    ):
        super(TorchVisionModel, self).__init__()
        if model_name == "squeezenet":
            self.model_ft = models.squeezenet1_0(pretrained=use_pretrained)
        else:
            value_error_log(
                logger, f"model_name: {model_name} is not supported!")
        BlocksTorch.set_parameter_requires_grad(self, feature_extract)
        self.model_ft.classifier[1] = nn.Conv2d(
            last_layer_nodes, num_classes, kernel_size=(1, 1), stride=(1, 1)
        )

    def forward(self, inputs):
        output = self.model_ft(inputs)

        return output



import numpy as np
import torch

from deep_utils.utils.utils.shuffle_utils import shuffle_group_torch


class CutMixTorch:
    @staticmethod
    def seg_cutmix_batch(
        a_images, a_masks, b_images=None, b_masks=None, beta=1, shuffle=True
    ):
        """
        Cutmix operation for two batch of segmentation images with their corresponding masks! In case of None input
        for `b_images` and `b_masks`, `a_images` and `b_images` are used instead of them.
        :param beta:
        :param a_images:
        :param a_masks:
        :param b_images:
        :param b_masks:
        :param shuffle:
        :return:
        """
        if b_images is None:
            b_images = a_images
            b_masks = a_masks

        a_images, a_masks, b_images, b_masks = (
            torch.clone(a_images),
            torch.clone(a_masks),
            torch.clone(b_images),
            torch.clone(b_masks),
        )

        if shuffle:
            shuffle_group_torch(a_images, a_masks)
            shuffle_group_torch(b_images, b_masks)

        x_cutmix, y_cutmix = [], []
        for a_img, b_img, a_mask, b_mask in zip(a_images, b_images, a_masks, b_masks):
            x, y = CutMixTorch.seg_cutmix(a_img, a_mask, b_img, b_mask, beta)
            x_cutmix.append(x)
            y_cutmix.append(y)
        return np.array(x_cutmix, dtype=np.uint8), np.array(y_cutmix, dtype=np.uint8)

    @staticmethod
    def seg_cutmix(a_img, a_mask, b_img, b_mask, beta=1):
        """
        Cutmix operation for two individual segmentation images with their corresponding masks!
        :param a_img:
        :param a_mask:
        :param b_img:
        :param b_mask:
        :param beta: alpha value for beta distribution, default is one as suggested by the paper!
        :return:
        """
        # Get lambda & box
        lam = np.random.beta(beta, beta)
        (x1, y1, x2, y2) = CutMixTorch.get_bbox(a_img.shape[:-1], lam)
        # generate mask
        mask = np.ones_like(a_img)
        mask[:, x1:x2, y1:y2] = 0
        # generate x
        x = (np.multiply(a_img, mask) + np.multiply(b_img, (abs(1.0 - mask)))).astype(
            np.uint8
        )
        # generate y
        mask = mask[0, :, :]
        y = (np.multiply(a_mask, mask) + np.multiply(b_mask, (abs(1.0 - mask)))).astype(
            np.uint8
        )
        return x, y

    @staticmethod
    def cls_cutmix_batch(
        a_images: torch.Tensor,
        a_labels: torch.Tensor,
        b_images: torch.Tensor = None,
        b_labels: torch.Tensor = None,
        beta=1,
        shuffle=True,
    ):
        """

        :param a_images:
        :param a_labels:
        :param b_images:
        :param b_labels:
        :param beta:
        :param shuffle:
        :return:
        """

        if b_images is None:
            b_images = a_images
            b_labels = a_labels

        a_images, a_labels, b_images, b_labels = (
            torch.clone(a_images),
            torch.clone(a_labels),
            torch.clone(b_images),
            torch.clone(b_labels),
        )

        if shuffle:
            a_images, a_labels = shuffle_group_torch(a_images, a_labels)
            b_images, b_labels = shuffle_group_torch(b_images, b_labels)

        x_cutmix, y_cutmix = [], []
        for a_img, b_img, a_label, b_label in zip(
            a_images, b_images, a_labels, b_labels
        ):
            img_cutmix, label_cutmix = CutMixTorch._cls_cutmix(
                a_img, a_label, b_img, b_label, beta
            )
            x_cutmix.append(img_cutmix.unsqueeze(0))
            y_cutmix.append(label_cutmix.unsqueeze(0))

        return torch.concat(x_cutmix, dim=0), torch.concat(y_cutmix, dim=0)

    @staticmethod
    def get_bbox(sizes, lam):
        """
        Get cutmix bboxes from input given sizes
        :param sizes:
        :param lam:
        :return:
        """
        if len(sizes) == 4:
            b = sizes[0]
            boxes = np.array([CutMixTorch._get_bbox(sizes[1:], lam)
                             for _ in range(b)])
        elif len(sizes) == 3:
            boxes = np.array(CutMixTorch._get_bbox(sizes, lam))
        else:
            raise ValueError(f"input size: {len(sizes)} is not valid!")
        return boxes

    @staticmethod
    def _get_bbox(size, lam):
        """
        Get random bbox from input size with the given lambda
        :param size:
        :param lam:
        :return:
        """
        w = size[1]
        h = size[2]

        cut_rat = np.sqrt(1.0 - lam)

        r_w = np.int(w * cut_rat)
        r_h = np.int(h * cut_rat)
        assert np.isclose((r_h * r_w) / (w * h), 1 - lam, atol=0.1)

        # uniform
        r_x = np.random.randint(w)
        r_y = np.random.randint(h)

        x1 = np.clip(r_x - r_w // 2, 0, w)
        y1 = np.clip(r_y - r_h // 2, 0, h)
        x2 = np.clip(r_x + r_w // 2, 0, w)
        y2 = np.clip(r_y + r_h // 2, 0, h)

        return x1, y1, x2, y2

    @staticmethod
    def _cls_cutmix(a_img, a_label, b_img, b_label, beta):
        assert (
            len(a_label.shape) == 1 and len(b_label.shape) == 1
        ), "label should be in one-hot format"
        lam = torch.tensor(np.random.beta(beta, beta))
        (x1, y1, x2, y2) = CutMixTorch.get_bbox(a_img.shape, lam)
        # create mask
        img_cutmix_mask = torch.ones_like(a_img)
        img_cutmix_mask[:, x1:x2, y1:y2] = 0
        img_cutmix = torch.multiply(a_img, img_cutmix_mask) + torch.multiply(
            b_img, (abs(1.0 - img_cutmix_mask))
        )
        label_cutmix = lam * a_label + b_label * (1 - lam)
        return img_cutmix, label_cutmix



from typing import List, Tuple, Union


class NEROperation:
    """
    This is a basic ner operation that validates ner-type in __init__ and implemented a simple __call__ method for the
    operation at hand.
    """

    NER_TYPES = ["BIO"]

    def __init__(self, ner_type="BIO"):
        self.ner_type = ner_type
        assert (
            ner_type in self.NER_TYPES
        ), f"[ERROR] ner_type: {self.ner_type} is not supported!"

    def __call__(
        self,
        tokens_x: Union[Tuple[str], List[str]],
        labels_y: Union[Tuple[str], List[str]],
    ):
        assert len(tokens_x) == len(
            labels_y
        ), f"[ERROR] number of tokens and labels is not equal"



import numpy as np

from deep_utils.main_abs.cv2.cv2_caffe import CV2Caffe
from deep_utils.utils.box_utils.boxes import Box
from deep_utils.utils.lib_utils.lib_decorators import (
    expand_input,
    get_elapsed_time,
    get_from_config,
    rgb2bgr,
)
from deep_utils.utils.resize_utils.main_resize import get_img_shape
from deep_utils.vision.face_detection.main import FaceDetector


class SSDCV2CaffeFaceDetector(FaceDetector):
    def __init__(self, **kwargs):
        super().__init__(name=self.__class__.__name__, file_path=__file__, **kwargs)

    def load_model(self):
        self.model = CV2Caffe(self.config)

    @rgb2bgr("bgr")
    @get_elapsed_time
    @expand_input(3)
    @get_from_config
    def detect_faces(
        self,
        img,
        is_rgb,
        resize_size=None,
        img_scale_factor=None,
        img_mean=None,
        resize_mode=None,
        confidence=None,
        get_time=False,
    ):
        faces = self.model.forward(
            img,
            resize_size=resize_size,
            img_scale_factor=img_scale_factor,
            img_mean=img_mean,
            resize_mode=resize_mode,
        )
        n_image, h, w, _ = get_img_shape(img)
        boxes = [[] for _ in range(n_image)]
        confidences = [[] for _ in range(n_image)]
        for i in range(faces.shape[2]):
            img_number = int(faces[0, 0, i, 0])
            face_confidence = faces[0, 0, i, 2]
            if face_confidence < confidence:
                continue
            box = faces[0, 0, i, 3:7] * np.array([w, h, w, h])
            box = Box.box2box(box, in_source="CV",
                              to_source="Numpy", return_int=True)
            boxes[img_number].append(box)
            confidences[img_number].append(face_confidence)
        return dict(boxes=boxes, confidences=confidences)


if __name__ == "__main__":
    model = SSDCV2CaffeFaceDetector()



try:
    from deep_utils.dummy_objects.vision.face_detection import RetinaFaceTorchFaceDetector
    from .torch.retinaface_torch_face_detection import RetinaFaceTorchFaceDetector
except ModuleNotFoundError:
    pass



from deep_utils.main_abs.main_config import MainConfig
import albumentations as A
from albumentations.pytorch import ToTensorV2


class Config(MainConfig):
    def __init__(self):
        self.normalizer = "l2_normalizer"
        self.img_w = 224
        self.img_h = 224
        self.device = "cpu"
        self.model_name = "senet50"
        self.model = None
        self.model_urls = {
            "senet50": "https://github.com/pooya-mohammadi/deep_utils/releases/download/0.9.5/senet50_ft_weight.pkl"
        }
        self.model_caches = {
            "senet50": "weights/vision/face_recognition/vggface2/torch/senet50/senet50_ft_weight.pkl"
        }
        self.download_variables = ("model",)

        self.model_means = {
            "senet50": [91.4953, 103.8827, 131.0912]
        }
        self.model_stds = {
            "senet50": [1, 1, 1]
        }

    @property
    def transform(self):
        return A.Compose(
            [A.Resize(self.img_w, self.img_h, always_apply=True),
             A.Normalize(self.model_mean, std=self.model_std, max_pixel_value=255.0, always_apply=True),
             ToTensorV2()
             ], p=1)

    @property
    def model_url(self):
        return self.model_urls[self.model_name]

    @property
    def model_cache(self):
        return self.model_caches[self.model_name]

    @property
    def model_mean(self):
        return self.model_means[self.model_name]

    @property
    def model_std(self):
        return self.model_stds[self.model_name]



from typing import List, Tuple, Union
import librosa
import numpy as np
import torch
from torch.nn import Module
from transformers import Wav2Vec2ForCTC, Wav2Vec2Processor
from deep_utils.audio.audio_utils.torchaudio_utils import TorchAudioUtils


class Wav2Vec2STTTorch:
    def __init__(self, model_path, device="cpu", sample_rate=16000, ctcdecoder=False, kenlm_model_path=None):
        self.device = device
        self.sample_rate = sample_rate
        self.processor = Wav2Vec2Processor.from_pretrained(model_path)
        self.model: Module = Wav2Vec2ForCTC.from_pretrained(
            model_path).eval().to(self.device)
        self.ctcdecoder = ctcdecoder
        if ctcdecoder:
            from pyctcdecode import build_ctcdecoder
            self.decoder = build_ctcdecoder(list(self.processor.tokenizer.get_vocab().keys()),
                                            kenlm_model_path=kenlm_model_path)

    def stt(self, speech_array, sr, mean=True) -> Union[str, torch.Tensor]:
        """

        :param speech_array:
        :param sr:
        :param mean: if set to True, get the mean of channels that are more than 1
        :return:
        """
        if len(speech_array.shape) == 2 and speech_array.shape[0] != 1:
            speech_array = torch.mean(speech_array, dim=0, keepdim=True)
        if len(speech_array.shape) == 2:
            speech_array = speech_array.squeeze(0)
        if sr != self.sample_rate:
            speech_array = TorchAudioUtils.resample(speech_array, sr, self.sample_rate)
        with torch.no_grad():
            input_values = self.processor(
                speech_array, sampling_rate=self.sample_rate, return_tensors="pt"
            ).input_values.to(self.device)
            logits = self.model(input_values).logits
            if self.ctcdecoder:
                logits = logits.cpu().numpy()[0]
                return self.decoder.decode(logits)
            else:
                pred_ids = torch.argmax(logits, dim=-1)
                pred_str = self.processor.batch_decode(pred_ids)[0]
        return pred_str

    def stt_file(self, file_path) -> Tuple[np.ndarray, float, str]:
        speech, sr = librosa.load(file_path, self.sample_rate)
        return speech, sr, self.stt(speech, sr)

    def stt_group(self, audio_segments, sr) -> List[str]:
        texts = []
        for audio in audio_segments:
            text = self.stt(audio, sr)
            texts.append(text)
        return texts



try:
    from deep_utils.dummy_objects.audio.classification.wav2vec2 import AudioClassificationWav2vec2Torch
    from .audio_classification_wav2vec2 import AudioClassificationWav2vec2Torch
except ModuleNotFoundError:
    pass



import pytest
import numpy as np
from deep_utils import TorchVisionUtils


@pytest.mark.torchvision
def test_detection():
    import albumentations as A
    from albumentations.pytorch import ToTensorV2
    from torchvision.transforms import transforms

    # albumentation test
    height, width, mean, std = 224, 220, 0, 1
    transform = A.Compose(
        [A.Resize(height, width),
         A.Normalize(mean, std, max_pixel_value=255.0),
         ToTensorV2()
         ])
    img_1 = np.random.randint(0, 255, (256, 180, 3), dtype=np.uint8)
    img_2 = np.random.randint(0, 255, (224, 289, 3), dtype=np.uint8)

    images = TorchVisionUtils.transform_concatenate_images([img_1, img_2], transform=transform)
    assert tuple(images.shape) == (
        2, 3, height, width), "TorchVisionUtils transform_concatenate_images not good for albumentation"

    # TorchVision test
    transform = transforms.Compose([
        transforms.Grayscale(),
        transforms.Resize((height, width)),
        transforms.ToTensor(),
        transforms.Normalize(mean=mean, std=std)]
    )
    images = TorchVisionUtils.transform_concatenate_images([img_1, img_2], transform=transform)
    assert tuple(images.shape) == (
        2, 1, height, width), "TorchVisionUtils transform_concatenate_images not good for torchvision"



from typing import Union, List

import numpy as np
import torch

from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.lib_utils.lib_decorators import (
    expand_input,
    get_elapsed_time,
    get_from_config,
    lib_rgb2bgr
)
from deep_utils.utils.pickle_utils.pickles import load_pickle
from deep_utils.vision.face_recognition.main.main_face_recognition import FaceRecognition, OUTPUT_CLASS
from .config import Config
from .src import load_model


class VggFace2TorchFaceRecognition(FaceRecognition):
    def __init__(self, **kwargs):
        super().__init__(
            name=self.__class__.__name__,
            file_path=__file__,
            download_variables=("model",),
            **kwargs
        )
        self.config: Config

    @download_decorator
    def load_model(self):
        # LOAD MODELS
        model = load_model(self.config.model_name)
        self.load_state_dict(model, self.config.model)
        model = model.eval()
        self.model = model

    @expand_input(3)
    @get_elapsed_time
    @get_from_config
    def extract_faces(self, img: Union[List[np.ndarray], np.ndarray], is_rgb, get_time=False) -> OUTPUT_CLASS:
        img = torch.cat(
            [self.config.transform(image=lib_rgb2bgr(img_, target_type="bgr", is_rgb=False))["image"].unsqueeze(0) for
             img_ in img], dim=0)

        with torch.no_grad():
            img = img.to(self.config.device)
            output = self.model(img).cpu().numpy()
            output = self.normalizer.transform(output)
        output = OUTPUT_CLASS(encodings=output, )
        return output

    @staticmethod
    def load_state_dict(model, weight_path):
        """
        Set parameters converted from Caffe models authors of VGGFace2 provide.
        See https://www.robots.ox.ac.uk/~vgg/data/vgg_face2/.

        Arguments:
            model: model
            weight_path: file name of parameters converted from a Caffe model, assuming the file format is Pickle.
        """
        weights = load_pickle(weight_path, mode="rb", encoding='latin1')
        own_state = model.state_dict()
        for name, param in weights.items():
            if name in own_state:
                try:
                    own_state[name].copy_(torch.from_numpy(param))
                except Exception:
                    raise RuntimeError(
                        'While copying the parameter named {}, whose dimensions in the model are {} and whose ' \
                        'dimensions in the checkpoint are {}.'.format(name, own_state[name].shape, param.shape))
            else:
                raise KeyError('unexpected key "{}" in state_dict'.format(name))



import torch
from PIL import Image


class TorchVisionUtils:
    @staticmethod
    def transform_concatenate_images(images, transform, mode=None, device='cpu'):
        """
        Concat and transform a batch images
        :param images:
        :param transform:
        :param mode:
        :param device:
        :return:
        """
        if mode == 'torchvision' or 'torchvision' in str(transform.__class__):
            output = torch.cat([transform(Image.fromarray(sample)).unsqueeze(0) for sample in images]).to(device)
        elif mode == "albumentation" or "albumentation" in str(transform.__class__):
            output = torch.cat([transform(image=sample)["image"].unsqueeze(0) for sample in images]).to(device)
        else:
            raise ValueError(f"mode: {mode} is not supported! [albumentation, torchvision]")
        return output



import pytest

from deep_utils import ndarray_to_b64, b64_to_ndarray


@pytest.mark.basic
def test_ndarray_to_b64_to_ndarray():
    """
    Testing ndarray_to_b64 and b64_to_ndarray
    Returns: Nothing

    """
    import numpy as np
    array = np.random.random((20, 5)).astype(np.float32)
    res = ndarray_to_b64(array, decode="utf-8")
    new_array = b64_to_ndarray(
        res, dtype="float32", shape=(20, 5), encode="utf-8")
    assert np.all(array == new_array), "with shapes and dtype does not work"

    res = ndarray_to_b64(array, append_shape=True, append_dtype=True)
    new_array = b64_to_ndarray(res, dtype=None, shape=None)
    assert np.all(array == new_array), "with appended shapes and dtype does not work"

    res = ndarray_to_b64(array, append_shape=True, append_dtype=True)
    new_array = b64_to_ndarray(res.decode(), dtype=None, shape=None)
    assert np.all(array == new_array), "with input str and appended shapes and dtype does not work"



import numpy as np
from deep_utils.utils.box_utils.boxes import Point
from deep_utils.utils.resize_utils.main_resize import resize


class VideoWriterCV:
    def __init__(
            self,
            save_path,
            width,
            height,
            fourcc="XVID",
            fps=30,
            colorful=True,
            in_source="Numpy",
    ):
        self.width, self.height = width, height
        import cv2

        point = Point.point2point(
            (width, height), in_source=in_source, to_source=Point.PointSource.CV
        )
        fourcc = cv2.VideoWriter_fourcc(
            *fourcc) if isinstance(fourcc, str) else fourcc
        self.vw = cv2.VideoWriter(save_path, fourcc, fps, point, colorful)

    def write(self, frame):
        if frame.shape[:2] != (self.width, self.height):
            frame = resize(frame, (self.width, self.height))
        self.vw.write(frame)


def rotate(
        img,
        rotation_degree,
        center_point=None,
        scale=1.0,
        dsize=None,
        bound=False,
        clockwise=True,
):
    import cv2

    h, w = img.shape[:2]
    (w, h) = dsize = (w, h) if dsize is None else dsize
    center_point = (w // 2, h // 2) if center_point is None else center_point
    # negative angle >> clockwise rotation | positive angle >> counter clockwise rotation
    rotation_degree = -rotation_degree if clockwise else rotation_degree
    m = cv2.getRotationMatrix2D(center_point, rotation_degree, scale)
    if bound:
        h, w = img.shape[:2]
        cos = abs(m[0, 0])
        sin = abs(m[0, 1])
        w_ = int((cos * w) + (sin * h))
        h_ = int((cos * h) + (sin * w))
        m[0, 2] += w_ // 2 - w // 2
        m[1, 2] += h_ // 2 - h // 2
        dsize = (w_, h_)
    rotated = cv2.warpAffine(img, m, dsize)

    return rotated


def translate(img, tx, ty, dsize=None):
    import cv2

    h, w = img.shape[:2][::-1]
    dsize = (w, h) if dsize is None else dsize
    translation_matrix = np.array([[1, 0, tx], [0, 1, ty]], dtype=np.float32)
    translated_image = cv2.warpAffine(
        src=img, M=translation_matrix, dsize=dsize)
    return translated_image


def show_destroy_cv2(img, win_name="", show=True):
    import cv2
    if show:
        try:
            cv2.imshow(win_name, img)
            cv2.waitKey(0)
            cv2.destroyWindow(win_name)
        except Exception as e:
            cv2.destroyWindow(win_name)
            raise e



try:
    from deep_utils.dummy_objects.vision.ocr.crnn.torch import CRNNModelTorch
    from .crnn_model import CRNNModelTorch
except:
    pass

try:
    from deep_utils.dummy_objects.vision.ocr.crnn.torch import CRNNInferenceTorch
    from .crnn_inference import CRNNInferenceTorch
except:
    pass



import os
import random
from typing import Dict, Union, List
from elasticsearch import Elasticsearch, NotFoundError
from deep_utils.utils.logging_utils.logging_utils import value_error_log
from deep_utils.utils.json_utils.json_utils import dump_json


class ElasticsearchEngin:
    def __init__(self, elastic_url="http://localhost:9200", es=None, timeout=10, logger=None, verbose=1):
        if es is None:
            self.es = Elasticsearch(elastic_url, timeout=timeout)
        elif isinstance(es, Elasticsearch):
            self.es = es
        else:
            value_error_log(logger, "url or es instance are not provided")
        self.verbose = verbose

    def get_bool_must_match_all_geo_filter(self, index_name, lat, lon, distance, scale="km", size=10):
        """
        GET index-name/_search
{
  "query": {
    "bool": {
      "must": {
        "match_all": {}
      },
      "filter": {
        "geo_distance": {
          "distance": "0.1km",
          "location": {
            "lat": 35.7041126,
            "lon": 51.4629228
          }
        }
      }
    }
  }
}
        :param index_name:
        :param lat:
        :param lon:
        :param distance:
        :param scale:
        :param size:
        :return:
        """
        query = {
            "bool": {
                "must": self.get_match_query(keyword="match_all"),
                "filter": self.get_geo_filter(lat, lon, distance, scale=scale)
            }
        }
        results = self.es.search(index=index_name, query=query, size=size)
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    @staticmethod
    def get_geo_filter(lat, lon, distance, scale="km"):
        filter_ = {
            "geo_distance": {
                "distance": f"{distance}{scale}",
                "location": {
                    "lat": lat,
                    "lon": lon
                }
            }
        }
        return filter_

    def get_suggestion_output(self, text, field_name, index_name="tehran-roads-centers"):
        query = {
            "match": {
                field_name: text
            }
        }
        suggest = {
            "my-suggestion": {
                "field_value": text,
                "term": {
                    "field": field_name
                }
            }
        }
        results = self.es.search(index=index_name, query=query, suggest=suggest)
        hits = results['hits']['hits']
        if len(hits) > 1:
            return text
        options = results['suggest']['my-suggestion'][0]['options']
        if len(options) == 0:
            # return the exact field_value
            return text
        else:
            return options[0]['field_value']

    @staticmethod
    def get_index_value_hits(hits, index, field_name):
        if not hits:
            return None
        else:
            return hits[index]['_source'].get(field_name, None)

    @staticmethod
    def _get_id(result: dict, area_type=False):
        hits = result["hits"]['hits']
        if len(hits) == 0:
            if area_type:
                return None, None
            return None
        if len(hits) > 1:
            value = hits[0]["_source"]
            province = value['province_name']
            city_name = value['name']
            os.makedirs("equals", exist_ok=True)
            dump_json(f"equals/{province}_{city_name}.json", hits, ensure_ascii=False)
            print(f"There are two objects with the same characteristics with the same name, {hits}")
            hits = [hits[0]]

        if len(hits) == 1:
            value = hits[0]["_source"]
            _id = value['id']
            if area_type:
                area_type = value['area_type']
                return _id, area_type
            else:
                return _id
        else:
            raise ValueError("We've got a problem")

    @staticmethod
    def term_search(es_client, field_term_dict: dict, index_name: str):

        query = ElasticsearchEngin.get_bool_must_constant_score_query(field_term_dict)

        results = es_client.search(index=index_name, query=query)
        return results.body

    @staticmethod
    def get_geo_sort(x, y, field_name="location", order="asc"):
        if x and y:
            geo_sort = {
                "_geo_distance": {
                    field_name: [x, y],
                    "order": order,
                    "unit": "km"
                }
            }
            sort = [geo_sort]
        else:
            sort = []

        return sort

    @staticmethod
    def get_sort_query(field_name, order="asc"):
        """
        This function creates a sort request
        :param field_name: The field to sort
        :param order: the order. Default is asc.
        :return:
        """
        if order and field_name:
            sort = {
                field_name: {
                    "order": order,
                }
            }
            sort = [sort]
        else:
            sort = []
        return sort

    @staticmethod
    def get_bool_must_constant_score_query(field_term_dict: dict):
        must_list = []
        for field, term in field_term_dict.items():
            if term is None:
                continue
            inner_query = {
                "constant_score": {
                    "filter": {
                        "term": {
                            field: term
                        }
                    }
                }
            }
            must_list.append(inner_query)

        query = {"bool": {"must": must_list}}
        return query

    @staticmethod
    def get_bool_must_fuzzy_query(field_term_dict: Dict[str, str]):
        """
        This method is used for getting a must query in fuzzy mode
        GET index-name/_search
        {
        "query":
            {"bool":
                {"must":
                    [{
                    "fuzzy":
                    {"field-name": {"value": "field-value"}}
                    }]
                }
            }
        }

        :param field_term_dict:
        :return:
        """
        must_list = ElasticsearchEngin.get_fuzzy_values(field_term_dict)

        query = {"bool": {"must": must_list}}
        return query

    @staticmethod
    def get_fuzzy_values(field_term_dict):
        """
        Returns the fuzzy values:
        [
         {"fuzzy": {"value_1": {"value": "value"}}},
         {"fuzzy": {"value_1": {"value": "value"}}}
        ]
        :param field_term_dict:
        :return:
        """
        must_list = []
        for field, term in field_term_dict.items():
            if term is None:
                continue
            inner_query = {
                "fuzzy": {
                    field: {
                        "value": term
                    }
                }
            }
            must_list.append(inner_query)
        return must_list

    def search_bool_must_fuzzy_query(self,
                                     field_term_dict: dict,
                                     index_name: str,
                                     size: Union[int, None] = None):
        """
        This method is used for searching a must query in fuzzy mode with size
        GET index-name/_search
        {
        "query":
            {"bool":
                {"must":
                    [{
                    "fuzzy":
                    {"field-name": {"value": "field-value"}}
                    }]
                }
            },
            "size": size
        }
        :param field_term_dict:
        :param index_name:
        :param size:
        :return:
        """
        query = self.get_bool_must_fuzzy_query(field_term_dict)
        results = self.es.search(index=index_name, query=query, size=size)
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    @staticmethod
    def get_query_multi_match_fuzzy(query_value: str,
                                    field_names: List[str],
                                    fuzziness: Union[int, str] = "AUTO"):
        """
        This method is used to get the query for multi_match which is used for searching a query value in various
        fields. The fuzziness is set to AUTO to compensate for spelling errors
        "query": {
            "multi_match" : {
            "type": best_fields,
              "query":    "query-val",
              "fields": [ "field-1", "field-2" ] ,
              "fuzziness": "AUTO"
            }
        }
        :param query_value:
        :param field_names:
        :param fuzziness: AUTO by default
        :return:
        """

        query = {
            "multi_match": {
                "query": query_value,
                "fields": field_names,
                "fuzziness": fuzziness,
                "type": "best_fields"
            }
        }
        return query

    def search_query_bool_must_multi_match_fuzzy(self,
                                                 index_name,
                                                 multi_match_query_value,
                                                 multi_match_field_names,
                                                 fuzzy_field_term_dict: dict,
                                                 size=None,
                                                 ):
        """
        GET index-name/_search
{
  "query": {
    "bool": {
      "must": [{
        "multi_match": {
          "query": multi_match_query_value,
          "type": "best_fields",
          "fields": multi_match_field_names,
          "fuzziness": "AUTO"
        }
      },
      {
        "fuzzy": {
          "value_1": {"value": "value"}
        }
      },
      {
        "fuzzy": {
          "value_2": {"value": "value"}
        }
      }
      ...
      ]
    }
  }
}
        :param index_name:
        :param multi_match_query_value:
        :param multi_match_field_names:
        :param fuzzy_field_term_dict:
        :param size:
        :return:
        """
        fuzzy_list = self.get_fuzzy_values(fuzzy_field_term_dict)
        multi_match = self.get_query_multi_match_fuzzy(multi_match_query_value, multi_match_field_names)
        query = {
            "bool": {
                "must": [multi_match, *fuzzy_list]
            }
        }
        results = self.es.search(index=index_name, query=query, size=size).body
        hits = self.get_hits(results)
        return hits

    def search_query_multi_match_fuzzy(self,
                                       index_name: str,
                                       query_value: str,
                                       field_names: List[str],
                                       fuzziness: Union[int, str] = "AUTO",
                                       size: Union[int, None] = None):
        """
        This method is used to get the query for multi_match which is used for searching a query value in various
        fields. The fuzziness is set to AUTO to compensate for spelling errors
        GET index-name/_search
        {
            "query": {
                "multi_match" : {
                  "query":    "query-val",
                  "fields": [ "field-1", "field-2" ] ,
                  "fuzziness": "AUTO"
                }
            },
            "size": size
        }
        :param index_name:
        :param query_value:
        :param field_names:
        :param fuzziness: set to AUTO
        :param size:
        :return:
        """
        query = self.get_query_multi_match_fuzzy(query_value=query_value, field_names=field_names, fuzziness=fuzziness)
        results = self.es.search(index=index_name, query=query, size=size)
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    @staticmethod
    def get_hits(results):
        """
        This is a simple method to extract hits or return none when the output of the search has no hits
        :param results:
        :return:
        """
        hits = results['hits']['hits']
        if len(hits) == 0:
            hits = None
        return hits

    def search_bool_must_fuzzy_query_geo_sort(self,
                                              field_term_dict: dict,
                                              index_name: str,
                                              lat: float,
                                              lon: float,
                                              geo_sort_field_name="location",
                                              geo_sort_order="asc",
                                              size: Union[int, None] = None):
        """
        This method is used for query-fuzzy search plus geo sort
        :param field_term_dict:
        :param index_name:
        :param lat:
        :param lon:
        :param geo_sort_field_name:
        :param geo_sort_order:
        :param size:
        :return:
        """
        assert lat is not None and lon is not None
        query = self.get_bool_must_fuzzy_query(field_term_dict)
        sort = self.get_geo_sort(lat, lon, field_name=geo_sort_field_name, order=geo_sort_order)
        results = self.es.search(index=index_name, query=query, sort=sort, size=size).body
        hits = self.get_hits(results)
        return hits

    def get_bool_must_constant_score_query_geo_sort(self,
                                                    field_term_dict: dict,
                                                    index_name: str,
                                                    lat: float, lon: float,
                                                    size: Union[int, None] = None):
        assert lat is not None and lon is not None
        query = self.get_bool_must_constant_score_query(field_term_dict)
        sort = self.get_geo_sort(lat, lon)
        results = self.es.search(index=index_name, query=query, sort=sort, size=size).body
        hits = self.get_hits(results)
        return hits

    def search_bool_must_fuzzy_query_sort(self,
                                          field_term_dict: dict,
                                          index_name: str,
                                          sort_filed: str,
                                          sort_order: str = "asc",
                                          size: Union[int, None] = None):
        """
        This method searches based on fuzzy based
        GET index_name/_search
{
  "query": {
    "bool": {
      "must": [
        {
          "fuzzy": {
            "file-name": {
              "value": "field-value"
            }
          }
        },
        {
          "fuzzy": {
            "file-name-2": {
              "value": "field-value-2"
            }
          }
        }
      ]
    }
  },
  "sort": [
    {
      "sort-field-name": {
        "order": "desc"
      }
    }
  ]
}
        :param field_term_dict:
        :param index_name:
        :param sort_filed:
        :param sort_order:
        :param size:
        :return:
        """
        assert sort_filed is not None and sort_order is not None
        query = self.get_bool_must_fuzzy_query(field_term_dict)
        sort = self.get_sort_query(field_name=sort_filed, order=sort_order)
        results = self.es.search(index=index_name, query=query, sort=sort, size=size).body
        hits = self.get_hits(results)
        return hits

    def search_bool_must_fuzzy_constant_score_range_sort(self,
                                                         field_term_dict: dict,
                                                         index_name: str,
                                                         sort_filed: str,
                                                         sort_order: str = "asc",
                                                         size: Union[int, None] = None):
        assert sort_filed is not None and sort_order is not None
        query = self.get_bool_must_constant_score_query(field_term_dict)
        sort = self.get_sort_query(field_name=sort_filed, order=sort_order)
        results = self.es.search(index=index_name, query=query, sort=sort, size=size).body
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    def get_match_all_geo_sort_query(self, index_name, lat, lon, sort_field="location", sort_order="asc", size=1):
        query = self.get_match_query(keyword="match_all")
        sort = self.get_geo_sort(lat, lon, field_name=sort_field, order=sort_order)
        results = self.es.search(index=index_name, query=query, sort=sort, size=size).body
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    @staticmethod
    def get_match_query(field_name="",
                        field_value="",
                        keyword="match",
                        ):
        """
        This function is used to get query-match. It will simply return query-match json
        :param field_name:
        :param field_value:
        :param keyword:
        :return:
        """
        if keyword == "match":

            query = {
                keyword: {
                    field_name: field_value
                }
            }
        elif keyword == "match_all":
            query = {
                keyword: {}
            }
        else:
            raise ValueError(f"keyword: {keyword} is not valid!")
        return query

    def search_match_query(self, index_name, field_name="", field_value="", keyword="match", size=None):
        query = self.get_match_query(field_name=field_name, field_value=field_value, keyword=keyword)
        results = self.es.search(index=index_name, query=query, size=size).body
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    def search_by_id(self, index_name, id_value, id_field_name="_id") -> Union[dict, None]:
        hits = self.search_match_query(index_name, field_name=id_field_name, field_value=id_value, keyword="match",
                                       size=1)
        if len(hits) == 1:
            return hits[0]["_source"]
        else:
            return None

    def search_query_match_sort(self, index_name, field_name, field_value, sort_field_name, sort_order, size=None):
        """
        This method is used to do a simple query match with sort.

        :param index_name:
        :param field_name:
        :param field_value:
        :param sort_field_name:
        :param sort_order:
        :param size:
        :return:
        """
        query_match = ElasticsearchEngin.get_match_query(field_name, field_value)
        sort = ElasticsearchEngin.get_sort_query(field_name=sort_field_name, order=sort_order)
        results = self.es.search(index=index_name, query=query_match, sort=sort, size=size).body
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    def search_query_match_geo_sort(self, index_name, field_name, field_value, geo_sort_field_name, geo_sort_order, lat,
                                    lon,
                                    size=None):
        """
        This method is used to do a simple query match with geo_sort.

        :param index_name:
        :param field_name:
        :param field_value:
        :param geo_sort_field_name:
        :param geo_sort_order:
        :param lat:
        :param lon:
        :param size:
        :return:
        """
        query_match = ElasticsearchEngin.get_match_query(field_name, field_value)
        geo_sort = ElasticsearchEngin.get_geo_sort(lat, lon, geo_sort_field_name, geo_sort_order)
        results = self.es.search(index=index_name, query=query_match, sort=geo_sort, size=size).body
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    def search_query_match_sort_geo_sort(self, index_name, field_name, field_value, geo_sort_field_name, geo_sort_order,
                                         sort_field_name, sort_order, lat, lon, size=None):
        """
        This method is used to do a simple query match with geo_sort plus a usual sort
        :param index_name:
        :param field_name:
        :param field_value:
        :param geo_sort_field_name:
        :param geo_sort_order:
        :param sort_field_name:
        :param sort_order:
        :param lat:
        :param lon:
        :param size:
        :return:
        """
        query_match = ElasticsearchEngin.get_match_query(field_name, field_value)
        geo_sort = ElasticsearchEngin.get_geo_sort(lat, lon, geo_sort_field_name, geo_sort_order)
        sort = ElasticsearchEngin.get_sort_query(field_name=sort_field_name, order=sort_order)
        results = self.es.search(index=index_name, query=query_match, sort=geo_sort + sort, size=size).body
        hits = ElasticsearchEngin.get_hits(results)
        return hits

    def update_by_id_add_new_field(self, index_name, id_, new_field_name, new_field_value, scripted_upsert=False,
                                   detect_noop=False):
        """
        This adds a new field to an existing document!
        POST test/_update/1
        {
          "script" : "ctx._source.new_field = 'value_of_new_field'"
        }


        :param index_name:
        :param id_:
        :param new_field_name:
        :param new_field_value:
        :param scripted_upsert:
        :param detect_noop:
        :return:
        """
        script = f"ctx._source.{new_field_name} = '{new_field_value}'"
        result = self.es.update(index=index_name,
                                id=id_,
                                detect_noop=detect_noop,
                                scripted_upsert=scripted_upsert,
                                script=script)
        return result

    def update_by_id_remove_field(self, index_name, id_, remove_field_name, scripted_upsert=False,
                                  detect_noop=False):
        """
        This adds a new field to an existing document!
        POST test/_update/1
        {
            "script" : "ctx._source.remove('new_field')"
        }

        :param index_name:
        :param id_:
        :param remove_field_name:
        :param scripted_upsert:
        :param detect_noop:
        :return:
        """
        script = f"ctx._source.remove('{remove_field_name}')"
        result = self.es.update(index=index_name,
                                id=id_,
                                detect_noop=detect_noop,
                                scripted_upsert=scripted_upsert,
                                script=script)
        return result

    def check_index_exists(self, index_name, create=False, es_mapping=Union[dict, None]):
        """
        Checks whether this index exists or not
        :param index_name:
        :param create: If set to true creates the index
        :param es_mapping:
        :return:
        """
        try:
            self.es.get(index=index_name, id="an id that might not be chosen by someone!")
        except NotFoundError as e:
            if e.message == "index_not_found_exception":
                if create:
                    self.es.indices.create(index=index_name, mappings=es_mapping)
                    # After being created return True:)
                    return True
                return False
        return True



import setuptools

VERSION = "0.10.7"

try:
    import pypandoc

    long_description = pypandoc.convert("Readme.md", "rst")
except (IOError, ImportError):
    long_description = open("Readme.md", mode="r", encoding="utf-8").read()

# Module dependencies
dependency_links = []

requirements = [
    "numpy>=1.21.0",
    "requests>=2.27.1",
    "tqdm>=4.62.3",
]

cv_requirements = ["opencv-python>=4.5.5.58"]

tf_requirements = [
                      "tensorflow>=2.6.0",
                  ] + cv_requirements

torch_requirements = [
    "torch>=1.8.0,<1.12.0",
    "torchvision>=0.10.0",
    "torchaudio>=1.10.0",
]

torchvision_requirements = [
                               "torch>=1.8.0,<1.12.0",
                               "torchvision>=0.10.0",
                           ] + cv_requirements
torch_transformers_requirements = torch_requirements + ["transformers>=4.18.0"]
setuptools.setup(
    name="deep_utils",
    version=VERSION,
    author="Pooya Mohammadi Kazaj",
    author_email="pooyamohammadikazaj@gmial.com",
    download_url=f"https://github.com/Practical-AI/deep_utils/archive/refs/tags/{VERSION}.tar.gz",
    description="Deep Utils",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/pooya-mohammadi/deep_utils",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    extras_require={
        "cv": cv_requirements,
        "tf": tf_requirements,
        "torch": torch_requirements,
        "torchvision": torchvision_requirements,
        "torch_transformers": torch_transformers_requirements,
    },
    install_requires=requirements,
    dependency_links=dependency_links,
    python_requires=">=3.6",
    license_files=('LICENSE',),
)



def check_num_exists(input_, lan="en"):
    """
    This function is used to check whether a character exists in 
    :param input_:
    :param lan:
    :return:
    """
    language = dict(en="0123456789", fa="۰۱۲۳۴۵۶۷۸۹")

    for num in language[lan]:
        if num in input_:
            return True
    return False



import sys

import cv2
import numpy as np

from deep_utils.utils.box_utils.boxes import Box, Point
from deep_utils.utils.lib_utils.download_utils import download_decorator
from deep_utils.utils.lib_utils.lib_decorators import (
    expand_input,
    get_elapsed_time,
    get_from_config,
    rgb2bgr,
)
from deep_utils.vision.face_detection.main.main_face_detection import FaceDetector

from .config import Config
from .utils.rfb_320 import create_rfb_net
from .utils.slim_320 import create_slim_net


class UltralightTFFaceDetector(FaceDetector):
    def __init__(self, **kwargs):
        super().__init__(
            name=self.__class__.__name__,
            file_path=__file__,
            download_variables=("RBF", "slim"),
            **kwargs
        )
        self.config: Config
        import tensorflow as tf

        self.tf = tf
        physical_devices = tf.config.experimental.list_physical_devices("GPU")
        if len(physical_devices) < 0:
            tf.config.experimental.set_memory_growth(physical_devices[0], True)

    @download_decorator
    def load_model(self):
        # LOAD MODELS
        RBF = create_rfb_net()
        slim = create_slim_net()

    @get_from_config
    @get_elapsed_time
    @rgb2bgr("rgb")
    @expand_input(3)
    def detect_faces(
        self,
        images,
        is_rgb,
        net_type,
        resize_size=None,
        img_scale_factor=None,
        img_mean=None,
        resize_mode=None,
        confidence=None,
        get_time=False,
    ):
        net_type = str.lower(net_type)
        if net_type == "slim":
            model_path = self.config.slim
        elif net_type == "rbf":
            model_path = self.config.RBF
        else:
            print("The net type is wrong!")
            sys.exit(1)

        model = self.tf.keras.models.load_model(model_path)
        if images.ndim < 4:
            images = np.array([images])

        boxes_, confidences_, landmarks_ = [], [], []
        for img_n in range(images.shape[0]):
            h, w, _ = images[img_n].shape
            img_resize = cv2.resize(images[img_n], (320, 240))
            img_resize = img_resize - 127.0
            img_resize = img_resize / 128.0
            img_resize = np.expand_dims(img_resize, axis=0)

            # result=[background,face,x1,y1,x2,y2]
            results = model.predict(img_resize)
            conf, box = [], []
            for result in results:
                start_x = int(result[2] * w)
                start_y = int(result[3] * h)
                end_x = int(result[4] * w)
                end_y = int(result[5] * h)
                boxes = [start_x, start_y, end_x, end_y]
                boxes = Box.box2box(
                    boxes, in_source=Box.BoxSource.CV, to_source=Box.BoxSource.Numpy
                )
                confidence_ = result[1]
                if confidence_ >= confidence:
                    conf.append(confidence_)
                    box.append(boxes)
            confidences_.append(conf)
            boxes_.append(box)
        return dict(boxes=boxes_, confidences=confidences_, landmarks=landmarks_)



from .dir_utils import (
    cp_mv_all,
    crawl_directory_dataset,
    dir_train_test_split,
    find_file,
    mkdir_incremental,
    remove_create,
    split_dir_of_dir,
    split_segmentation_dirs,
    split_xy_dir,
    transfer_directory_items,
    combine_directory_of_directories,
    file_incremental
)



from pathlib import Path
from typing import Union

from deep_utils.utils.logging_utils import log_print


def load_yaml(yaml_path: Union[Path, str], logger=None, verbose=1) -> dict:
    """
    Load a yaml file to a dictionary object
    Args:
        yaml_path:
        logger:
        verbose:

    Returns:

    """
    import yaml

    with open(yaml_path, mode='r') as f:
        data_map = yaml.safe_load(f)
    log_print(logger, f"Successfully loaded {yaml_path}", verbose=verbose)
    return data_map


def dump_yaml(
    yaml_dict: dict, yaml_path: Union[Path, str], logger=None, verbose=1
) -> None:
    """
    Dumps a dict to yaml file
    Args:
        yaml_dict:
        yaml_path:
        logger:
        verbose:

    Returns:

    """
    import yaml

    with open(yaml_path, "w") as outfile:
        yaml.dump(yaml_dict, outfile, default_flow_style=False)
    log_print(
        logger, f"Successfully converted input dict to {yaml_path}", verbose=verbose
    )


def yaml_post_process(yaml_dict: dict):
    res = dict()
    for k, v in yaml_dict.items():
        if isinstance(v, dict):
            res[k] = yaml_post_process(v)
        elif v in ["none", "None"]:
            res[k] = None
        else:
            res[k] = v
    return res



from typing import Dict


class PostgresqlUtils:
    @staticmethod
    def connect(connection=None, username=None, password=None, host=None, port=None):
        import psycopg2
        if connection is None:
            connection = psycopg2.connect(
                database="postgres", user=username, password=password, host=host, port=port)
        return connection

    @staticmethod
    def create_db(db_name, connection=None, username=None, password=None, host=None, port=None):
        from psycopg2.errors import DuplicateDatabase
        connection = PostgresqlUtils.connect(connection, username, password, host, port)
        connection.autocommit = True
        # Creating a cursor object using the cursor() method
        cursor = connection.cursor()
        sql = f'''CREATE database {db_name}'''
        try:
            cursor.execute(sql)
            print("Database created successfully........")
        except DuplicateDatabase:
            print("Database is already created!")
        connection.close()

    @staticmethod
    def check_connection(connection=None, username=None, password=None, host=None, port=None,
                         status_key="POSTGRES_STATUS") -> Dict[str, str]:
        connection = PostgresqlUtils.connect(connection, username, password, host, port)
        # check postgres
        output = {}
        try:
            postgres_status = connection.status
            output[status_key] = "Alive" if postgres_status else "Down"
        except:
            output[status_key] = "Down"
        connection.close()
        return output






from .arg_parser import easy_argparse
from .hyper_parameters import KeyValStruct, YamlConfig, yaml_config2yaml_file
from .seeds import tf_set_seed
from .shuffle_utils import shuffle_group, shuffle_group_torch
from .str_utils import color_str
from .yaml_utils import dump_yaml, load_yaml, yaml_post_process



try:
    from deep_utils.dummy_objects.utils.sqlalchemy_utils import SQLAlchemyChecks
    from .checks import SQLAlchemyChecks
except ModuleNotFoundError:
    pass

try:
    from deep_utils.dummy_objects.utils.sqlalchemy_utils import SQLAlchemyInserts
    from .insert import SQLAlchemyInserts
except ModuleNotFoundError:
    pass

try:
    from deep_utils.dummy_objects.utils.sqlalchemy_utils import SQLAlchemyUtils
    from .sqlalchemy_utils import SQLAlchemyUtils
except ModuleNotFoundError:
    pass



from typing import Union
from deep_utils.utils.re_utils.re_utils import REUtils


def persian_num2english(input_string: str, reverse: bool = False):
    """
    Converts persian numbers to english
    Args:
        input_string:
        reverse: If set to True, converts english 2 persian!

    Returns:

    """
    NUM_MAP = {
        "۱": "1",
        "۲": "2",
        "۳": "3",
        "۴": "4",
        "۵": "5",
        "۶": "6",
        "۷": "7",
        "۸": "8",
        "۹": "9",
        "۰": "0",
    }
    if reverse:
        NUM_MAP = {v: k for k, v in NUM_MAP.items()}
    output_string = "".join([NUM_MAP.get(c, c) for c in input_string])
    return output_string


def arabic_char2fa_char(input_string: str):
    arabic2persian = {
        "ك": "ک",
        "ي": "ی",
    }
    out_string = "".join(arabic2persian.get(s, s) for s in input_string)
    return out_string


def num2fa_spoken(num_string: Union[str, int], split_num_char: bool = False):
    """
    converting string number to the spoken format. This function converts both persian and english workds to persian
    spoken words
    :param num_string:
    :param split_num_char: if set to True, splits connected numbers and characters
    :return:
    >>> num2fa_spoken("30")
    'سی ام'
    >>> num2fa_spoken("21")
    'بیست و یکم'
    >>> num2fa_spoken("۳۲")
    'سی و دوم'
    >>> num2fa_spoken(2)
    'دوم'
    >>> num2fa_spoken("204غربی", split_num_char=True)
    'دویست و چهارم غربی'
    """

    num_string = str(num_string)
    if split_num_char:
        num_string = REUtils.split_char_number(num_string)
    num_string = " ".join([_num2fa_spoken(word) for word in num_string.split(" ")])
    return num_string


def _num2fa_spoken(num_string):
    from num2fawords import words
    if num_string.isdigit():
        spoken_num = words(num_string)
        if spoken_num[-2:] == "سه":
            spoken_num = spoken_num[:-2] + "سوم"
        elif spoken_num[-1] in ["ی", ]:
            spoken_num += " ام"
        else:
            spoken_num += "م"
    else:
        spoken_num = num_string
    return spoken_num


def num2fa_spoken_sentence(sentence: str, split_num_char=False) -> str:
    """
    Applies num2fa_spoken on a sentence of words!
    :param sentence:
    :param split_num_char: if set to True, splits connected numbers and characters
    :return:
    """
    return " ".join([num2fa_spoken(word, split_num_char=split_num_char) for word in sentence.split(" ")])



try:
    from deep_utils.dummy_objects.vision.object_detection import YOLOV5TorchObjectDetector
    from deep_utils.vision.object_detection.yolo.v5.torch.yolo_v5_torch_object_detection import (
        YOLOV5TorchObjectDetector,
    )
except ModuleNotFoundError:
    pass
except Exception as e:
    raise e



from abc import abstractmethod
from typing import Union

import numpy as np

from deep_utils.main_abs.main import MainClass


class ObjectDetector(MainClass):
    def __init__(self, name, file_path, *args, **kwargs):
        super().__init__(name, file_path=file_path, **kwargs)
        self.confidences: Union[list, None, np.ndarray] = None
        self.boxes: Union[list, None, np.ndarray] = None
        self.classes: Union[list, None, np.ndarray] = None
        self.class_names: Union[list, None, np.ndarray] = None

    @abstractmethod
    def detect_objects(
        self,
        img,
        is_rgb,
        confidence=None,
        iou_thresh=None,
        classes=None,
        get_time=False,
        **kwargs
    ):
        pass

    @abstractmethod
    def detect_dir(
        self,
        dir_,
        confidence=None,
        iou_thresh=None,
        classes=None,
        extensions=(".png", ".jpg", ".jpeg"),
        save_results=False,
        save_in_file=False,
        **kwargs
    ):
        pass



try:
    from deep_utils.dummy_objects.vision.face_detection import UltralightTorchFaceDetector
    from .ultralight_torch_face_detection import UltralightTorchFaceDetector
except ModuleNotFoundError:
    pass



from collections import defaultdict

import numpy as np


class CTCDecoder:
    NINF = -1 * float("inf")
    DEFAULT_EMISSION_THRESHOLD = 0.01

    @staticmethod
    def merge_duplicates_remove_blanks(labels, blank_class=0):
        if len(labels.shape) == 2:
            return [
                CTCDecoder._merge_duplicates_remove_blanks(blank_class, label)
                for label in labels
            ]
        else:
            return CTCDecoder._merge_duplicates_remove_blanks(blank_class, labels)

    @staticmethod
    def _merge_duplicates_remove_blanks(blank_class, labels):
        new_labels = []
        # merge duplicate labels
        previous = None
        for label in labels:
            if label != previous:
                new_labels.append(label)
                previous = label
        # remove blanks
        new_labels = [label for label in new_labels if label != blank_class]
        return new_labels

    @staticmethod
    def greedy_decode(log_prob: np.ndarray, blank_class=0, **kwargs):
        """
        Gets greedy samples
        :param log_prob:
        :param blank_class:
        :return:
        """
        labels = np.argmax(log_prob, axis=-1)
        labels = CTCDecoder.merge_duplicates_remove_blanks(
            labels, blank_class=blank_class
        )
        return labels

    @staticmethod
    def beam_search_decode(emission_log_prob, blank_class=0, **kwargs):
        from scipy.special import (
            logsumexp,
        )  # log(p1 + p2) = logsumexp([log_p1, log_p2])

        beam_size = kwargs["beam_size"]
        emission_threshold = kwargs.get(
            "emission_threshold", np.log(CTCDecoder.DEFAULT_EMISSION_THRESHOLD)
        )

        length, class_count = emission_log_prob.shape

        beams = [([], 0)]  # (prefix, accumulated_log_prob)
        for t in range(length):
            new_beams = []
            for prefix, accumulated_log_prob in beams:
                for c in range(class_count):
                    log_prob = emission_log_prob[t, c]
                    if log_prob < emission_threshold:
                        continue
                    new_prefix = prefix + [c]
                    # log(p1 * p2) = log_p1 + log_p2
                    new_accu_log_prob = accumulated_log_prob + log_prob
                    new_beams.append((new_prefix, new_accu_log_prob))

            # sorted by accumulated_log_prob
            new_beams.sort(key=lambda x: x[1], reverse=True)
            beams = new_beams[:beam_size]

        # sum up beams to produce labels
        total_accu_log_prob = {}
        for prefix, accu_log_prob in beams:
            labels = tuple(
                CTCDecoder.merge_duplicates_remove_blanks(
                    prefix, blank_class=blank_class
                )
            )
            # log(p1 + p2) = logsumexp([log_p1, log_p2])
            total_accu_log_prob[labels] = logsumexp(
                [accu_log_prob, total_accu_log_prob.get(
                    labels, CTCDecoder.NINF)]
            )

        labels_beams = [
            (list(labels), accu_log_prob)
            for labels, accu_log_prob in total_accu_log_prob.items()
        ]
        labels_beams.sort(key=lambda x: x[1], reverse=True)
        labels = labels_beams[0][0]

        return labels

    @staticmethod
    def prefix_beam_decode(emission_log_prob, blank=0, **kwargs):
        from scipy.special import (
            logsumexp,
        )  # log(p1 + p2) = logsumexp([log_p1, log_p2])

        beam_size = kwargs["beam_size"]
        emission_threshold = kwargs.get(
            "emission_threshold", np.log(CTCDecoder.DEFAULT_EMISSION_THRESHOLD)
        )

        length, class_count = emission_log_prob.shape

        beams = [
            (tuple(), (0, CTCDecoder.NINF))
        ]  # (prefix, (blank_log_prob, non_blank_log_prob))
        # initial of beams: (empty_str, (log(1.0), log(0.0)))

        for t in range(length):
            new_beams_dict = defaultdict(
                lambda: (CTCDecoder.NINF, CTCDecoder.NINF)
            )  # log(0.0) = NINF

            for prefix, (lp_b, lp_nb) in beams:
                for c in range(class_count):
                    log_prob = emission_log_prob[t, c]
                    if log_prob < emission_threshold:
                        continue

                    end_t = prefix[-1] if prefix else None

                    # if new_prefix == prefix
                    new_lp_b, new_lp_nb = new_beams_dict[prefix]

                    if c == blank:
                        new_beams_dict[prefix] = (
                            logsumexp(
                                [new_lp_b, lp_b + log_prob, lp_nb + log_prob]),
                            new_lp_nb,
                        )
                        continue
                    if c == end_t:
                        new_beams_dict[prefix] = (
                            new_lp_b,
                            logsumexp([new_lp_nb, lp_nb + log_prob]),
                        )

                    # if new_prefix == prefix + (c,)
                    new_prefix = prefix + (c,)
                    new_lp_b, new_lp_nb = new_beams_dict[new_prefix]

                    if c != end_t:
                        new_beams_dict[new_prefix] = (
                            new_lp_b,
                            logsumexp(
                                [new_lp_nb, lp_b + log_prob, lp_nb + log_prob]),
                        )
                    else:
                        new_beams_dict[new_prefix] = (
                            new_lp_b,
                            logsumexp([new_lp_nb, lp_b + log_prob]),
                        )

            # sorted by log(blank_prob + non_blank_prob)
            beams = sorted(
                new_beams_dict.items(), key=lambda x: logsumexp(x[1]), reverse=True
            )
            beams = beams[:beam_size]

        labels = list(beams[0][0])
        return labels

    @staticmethod
    def ctc_decode_batch(
        log_probs: np.ndarray,
        label2char=None,
        blank_class=0,
        decoder_name="greedy",
        beam_size=10,
    ):
        """
        Input array should be [batch, length, class]
        :param log_probs:
        :param label2char:
        :param blank_class:
        :param decoder_name:
        :param beam_size:
        :return:
        """

        decoded_list = [
            CTCDecoder.ctc_decode(
                log_prob, decoder_name, blank_class, label2char, beam_size
            )
            for log_prob in log_probs
        ]
        return decoded_list

    @staticmethod
    def ctc_decode(
        log_prob: np.ndarray,
        decoder_name="greedy",
        blank_class=0,
        label2char=None,
        beam_size=10,
    ):
        decoder = {
            "greedy": CTCDecoder.greedy_decode,
            "beam_search": CTCDecoder.beam_search_decode,
            "prefix_beam_search": CTCDecoder.prefix_beam_decode,
        }[decoder_name]
        decoded = decoder(log_prob, blank_class=blank_class,
                          beam_size=beam_size)
        if label2char:
            decoded = CTCDecoder.single_decode(decoded, label2char)
        return decoded

    @staticmethod
    def single_decode(prediction: list, label2char: dict):
        return [label2char[p] for p in prediction]

    @staticmethod
    def decode(predictions: list, label2char: dict):
        if isinstance(predictions[0], list):
            return [
                CTCDecoder.single_decode(prediction, label2char)
                for prediction in predictions
            ]
        else:
            return CTCDecoder.single_decode(predictions, label2char)



try:
    from deep_utils.dummy_objects.vision.torch_vision import TorchVisionInference
    from deep_utils.vision.torch_vision.torch_vision_inference.torch_vision_inference import (
        TorchVisionInference,
    )
except:
    pass



from .algorithm_utils import *
from .box_utils import *
from .coco_utils import *
from .compress_utils import *
from .ctc_decoder import *
from .decorators import *
from .dir_utils import *
from .encodes import *
from .image_utils import *
from .lib_utils import *
from .logging_utils import *
from .lr_scheduler_utils import *
from .matplotlib_utils import *
from .memory_utils import *
from .minio_lib import *
from .multi_label_utils import *
from .np_utils import *
from .opencv_utils import *
from .optimizers import *
from .os_utils import *
from .pickle_utils import *
from .random_utils import *
from .re_utils import *
from .resize_utils import *
from .tensorboard_utils import *
from .tf_utils import *
from .color_utils import *
from .gif_utils import *
from .dict_named_tuple_utils import *
from .dict_utils import *
from .list_utils import *
from .object_utils import *
from .torch_utils import *
from .shutil_utils import *
from .json_utils import *
from .requests_utils import *
from .postgresql_utils import *
from .kafka_utils import *
from .download_utils import *
from .hash_utils import *
from .variable_naming_utils import *
from .sqlalchemy import *
from .utils import *



from argparse import Namespace
from typing import Union


def get_obj_variables(obj) -> dict:
    """
    Gets the attributes of an object not the methods.
    :param obj:
    :return:
    """
    out = dict()
    for key in dir(obj):
        val = getattr(obj, key)
        if (key.startswith("__") and key.endswith("__")) or type(val).__name__ == "method":
            continue
        else:
            out[key] = val
    return out


def get_attributes(obj) -> dict:
    """
    Gets the attributes of an object not the methods. Deprecated: The naming was not appropriate, make sure to use
    `get_variables` instead.
    :param obj:
    :return:
    """
    return get_obj_variables(obj)


def update_obj_params(obj_instance: object, args: Union[dict, Namespace]):
    """
    This func/method is used for updating parameters of a class especially a config class with input arguments
    :param obj_instance:
    :param args:
    :return:
    """
    if isinstance(args, Namespace):
        variables = vars(args)
    elif isinstance(args, dict):
        variables = args
    else:
        raise ValueError()
    for k, v in variables.items():
        if hasattr(obj_instance, k):
            setattr(obj_instance, k, v)
        else:
            raise ValueError(f"value {k} is not defined in {obj_instance.__class__.__name__}")


def variable_repr(self):
    """
    A representation func for objects
    :param self:
    :return:
    """
    variables = get_obj_variables(self)
    return f"{self.__class__.__name__} -> " + ", ".join(f"{k}: {v}" for k, v in variables.items())



import torch
import torch.nn as nn
import torch.nn.functional as F
from deep_utils.blocks.torch.blocks_torch import BlocksTorch


class CRNNModelTorch(nn.Module):

    def __init__(self, img_h, n_channels, n_classes, n_hidden, lstm_input=64):
        super(CRNNModelTorch, self).__init__()
        assert img_h % 16 == 0, 'img_h has to be a multiple of 16'

        block_0 = BlocksTorch.conv_norm_act(n_channels, 64, pooling='max')
        block_1 = BlocksTorch.conv_norm_act(64, 128, pooling='max')
        block_2 = BlocksTorch.conv_norm_act(128, 256)
        block_3 = BlocksTorch.conv_norm_act(256, 256, pooling='max', pooling_s=(2, 1), pooling_k=(2, 1))
        block_4 = BlocksTorch.conv_norm_act(256, 512, norm="bn")
        block_5 = BlocksTorch.conv_norm_act(512, 512, pooling='max', norm="bn", pooling_s=(2, 1), pooling_k=(2, 1))
        block_6 = BlocksTorch.conv_norm_act(512, 512, k=2, p=0)
        self.cnn = nn.Sequential(block_0, block_1, block_2, block_3, block_4, block_5, block_6)
        self.cnn2rnn = nn.Linear(512 * (img_h // 16 - 1), lstm_input)

        # RNN
        self.rnn_01 = nn.LSTM(lstm_input, n_hidden, bidirectional=True)
        self.rnn_02 = nn.LSTM(2 * n_hidden, n_hidden, bidirectional=True)

        # Classifiers
        self.classifier = nn.Linear(2 * n_hidden, n_classes)

    @staticmethod
    def weights_init(m):
        classname = m.__class__.__name__
        if classname.find('Conv') != -1:
            m.weight.data.normal_(0.0, 0.02)
        elif classname.find('BatchNorm') != -1:
            m.weight.data.normal_(1.0, 0.02)
            m.bias.data.fill_(0)
        elif classname.find('Linear') != -1:
            torch.nn.init.xavier_uniform(m.weight)
            m.bias.data.fill_(0.01)

    def forward(self, input, return_cls=False):
        # conv features
        conv = self.cnn(input)
        batch_size, channels, h, w = conv.shape
        conv = conv.view((batch_size, channels * h, w)).permute(2, 0, 1)
        rnn_input = self.cnn2rnn(conv)
        # rnn features
        x, _ = self.rnn_01(rnn_input)
        rnn_output, _ = self.rnn_02(x)

        # Classifier
        cls = self.classifier(rnn_output)
        # add log_softmax to converge output
        output = F.log_softmax(cls, dim=2)
        if return_cls:
            return output, F.softmax(cls, 2)

        return output



import cv2
import pytest

from deep_utils import Box, MTCNNTorchFaceDetector, download_file


@pytest.mark.torchvision
def test_face_detection():
    model = MTCNNTorchFaceDetector(device="cpu")
    file_path = download_file(
        "https://github.com/pooya-mohammadi/deep-utils-notebooks/releases/download/0.1.0/movie-stars.jpg"
    )
    img = cv2.imread(file_path)
    res = model.detect_faces(img, is_rgb=False, get_time=True)
    assert len(res.boxes) >= 0, "Model did not detect anything!"
    img = Box.put_box(img, res.boxes)
    assert img is not None, "put_box returned nothing!"



try:
    from deep_utils.dummy_objects.vision.face_detection import UltralightTFFaceDetector
    from .ultralight_tf_face_detection import UltralightTFFaceDetector
except ModuleNotFoundError:
    pass



import importlib.util
from collections import OrderedDict
from deep_utils.utils.utils.str_utils import color_str


def is_available(module_name: str, lib_name: str = None):
    lib_name = module_name if lib_name is None else lib_name
    import_error = (
            "{0} "
            + f"""
    requires {module_name} library but it was not found in your environment. You can install it with the following instructions:
    ```
    pip install {lib_name}
    ```
    In a notebook or a colab, you can install it by executing a cell with
    ```
    !pip install {lib_name}
    ```
    """
    )
    return module_name, (
        lambda: importlib.util.find_spec(module_name) is not None,
        import_error,
    )


_is_tf_available = importlib.util.find_spec("tensorflow") is not None
_is_torch_available = importlib.util.find_spec("torch") is not None
_is_cv2_available = importlib.util.find_spec("cv2") is not None
_is_torchvision_available = importlib.util.find_spec("torchvision") is not None
_is_torchaudio_available = importlib.util.find_spec("torchaudio") is not None
_is_pyannote_audio_available = importlib.util.find_spec("pyannote") is not None


def is_torch_available():
    return _is_torch_available


def is_torchvision_available():
    return _is_torchvision_available


def is_torchaudio_available():
    return _is_torchaudio_available


def is_tf_available():
    return _is_tf_available


def is_cv2_available():
    return _is_cv2_available


def is_pyannote_audio_available():
    return _is_pyannote_audio_available


CV2_IMPORT_ERROR = """
{0} requires OpenCV library but it was not found in your environment. You can install it with:
```
pip install opencv-python
# or
pip install deep-utils[cv]
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install opencv-python
or
!pip install deep-utils[cv]
```
"""

TENSORFLOW_IMPORT_ERROR = """
{0} requires TensorFlow library but it was not found in your environment. You can install it with the following instruction or check out the main webpage of tensorflow.org:
```
pip install tensorflow
# or
pip install deep-utils[tf]
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install tensorflow
or
!pip install deep-utils[tf]
```
"""

PYTORCH_IMPORT_ERROR = """
{0} requires PyTorch library but it was not found in your environment. You can install it with the following instruction or check out the main webpage of pytorch.org:
```
pip install torch
# or
pip install deep-utils[torch]
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install torch
or
!pip install deep-utils[torch]
```
"""

TORCHVISION_IMPORT_ERROR = """
{0} requires Torchvision library but it was not found in your environment. You can install it with the following instruction or check out the main webpage of pytorch.org:
```
pip install torchvision
# or
pip install deep-utils[torchvision]
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install torchvision
or
!pip install deep-utils[torchvision]
```
"""

TORCHAUDIO_IMPORT_ERROR = """
{0} requires Torchaudio library, but it was not found in your environment. You can install it with the following instruction or check out the main webpage of pytorch.org:
```
pip install torchaudio
# or
pip install deep-utils[torch]
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install torchaudio
or
!pip install deep-utils[torch]
```
"""

PYANNOTE_AUDIO_IMPORT_ERROR = """
{0} requires PYANNOTE_AUDIO library, but it was not found in your environment. You can install it with the following instruction or check out the main repository page of pyannote(https://github.com/pyannote/pyannote-audio):
```
pip install pyannote.audio
```
In a notebook or a colab, you can install it by executing a cell with
```
!pip install pyannote.audio
```
"""

BACKENDS_MAPPING = OrderedDict(
    [
        ("cv2", (is_cv2_available, CV2_IMPORT_ERROR)),
        ("tf", (is_tf_available, TENSORFLOW_IMPORT_ERROR)),
        ("torch", (is_torch_available, PYTORCH_IMPORT_ERROR)),
        ("torchvision", (is_torchvision_available, TORCHVISION_IMPORT_ERROR)),
        ("torchaudio", (is_torchaudio_available, TORCHAUDIO_IMPORT_ERROR)),
        is_available("seaborn"),
        is_available("numpy"),
        is_available("albumentations"),
        is_available("sklearn"),
        is_available("PIL"),
        is_available("pyannote"),
        is_available("librosa"),
        is_available("transformers"),
        is_available("soundfile"),
        is_available("psutil"),
        is_available("yaml", "pyyaml"),
    ]
)


def requires_backends(obj, backends):
    if not isinstance(backends, (list, tuple)):
        backends = [backends]

    name = obj.__name__ if hasattr(obj, "__name__") else obj.__class__.__name__
    checks = (BACKENDS_MAPPING[backend] for backend in backends)
    failed = [msg.format(name) for available, msg in checks if not available()]
    if failed:
        color_str("".join(failed), color="red", mode=["bold", "underline"])
        print("".join(failed))


class DummyObject(type):
    """
    Metaclass for the dummy objects. Any class inheriting from it will return the ImportError generated by
    `requires_backend` each time a user tries to access any method of that class.
    """

    def __getattr__(cls, key):
        cls()



import cv2
import numpy as np

palette = (2**11 - 1, 2**15 - 1, 2**20 - 1)


def compute_color_for_labels(label):
    """
    Simple function that adds fixed color depending on the class
    """
    color = [int((p * (label**2 - label + 1)) % 255) for p in palette]
    return tuple(color)


def draw_boxes(img, bbox, identities=None, offset=(0, 0)):
    for i, box in enumerate(bbox):
        x1, y1, x2, y2 = [int(i) for i in box]
        x1 += offset[0]
        x2 += offset[0]
        y1 += offset[1]
        y2 += offset[1]
        # box text and bar
        id = int(identities[i]) if identities is not None else 0
        color = compute_color_for_labels(id)
        label = "{}{:d}".format("", id)
        t_size = cv2.getTextSize(label, cv2.FONT_HERSHEY_PLAIN, 2, 2)[0]
        cv2.rectangle(img, (x1, y1), (x2, y2), color, 3)
        cv2.rectangle(
            img, (x1, y1), (x1 + t_size[0] + 3, y1 + t_size[1] + 4), color, -1
        )
        cv2.putText(
            img,
            label,
            (x1, y1 + t_size[1] + 4),
            cv2.FONT_HERSHEY_PLAIN,
            2,
            [255, 255, 255],
            2,
        )
    return img


if __name__ == "__main__":
    for i in range(82):
        print(compute_color_for_labels(i))



from deep_utils.utils.lib_utils.main_utils import import_module

HaarcascadeCV2FaceDetector = import_module(
    "deep_utils.vision.face_detection.haarcascade.cv2_.haarcascade_cv2_face_detection",
    "HaarcascadeCV2FaceDetector",
)



class KafkaUtils:
    @staticmethod
    def create_topic(*topics, admin_client=None, bootstrap_servers="localhost:9092", num_partitions=1,
                     replication_factor=1, validate_only=False, client_id="kafka_utils",
                     logger=None, verbose=1):
        """
        Simply creates input topics
        :param topics:
        :param admin_client:
        :param bootstrap_servers:
        :param num_partitions:
        :param replication_factor:
        :param validate_only:
        :param client_id:
        :param logger:
        :param verbose:
        :return:
        """
        from kafka.admin import NewTopic
        from kafka.errors import TopicAlreadyExistsError
        from deep_utils.utils.logging_utils.logging_utils import log_print
        admin_client = KafkaUtils.create_admin_client(admin_client, bootstrap_servers, client_id)
        topic_list = [NewTopic(name=topic, num_partitions=num_partitions, replication_factor=replication_factor) for
                      topic in topics]
        try:
            admin_client.create_topics(new_topics=topic_list, validate_only=validate_only)
            log_print(logger, f"Successfully created {topics}", verbose=verbose)
        except TopicAlreadyExistsError:
            log_print(logger, f"Topics: {topics} already exist", verbose=verbose)

    @staticmethod
    def create_admin_client(admin_client=None, bootstrap_servers="localhost:9092", client_id="kafka_utils"):
        if admin_client is None:
            from kafka.admin import KafkaAdminClient
            admin_client = KafkaAdminClient(
                bootstrap_servers=bootstrap_servers,
                client_id=client_id,
                api_version=(0, 9)
            )
        return admin_client

    @staticmethod
    def create_producer(bootstrap_servers):
        from kafka import KafkaProducer
        producer = KafkaProducer(bootstrap_servers=bootstrap_servers)
        return producer

    @staticmethod
    def check_kafka_status(admin_client=None, bootstrap_servers="localhost:9092", status_key="KAFKA_STATUS"):
        admin_client = KafkaUtils.create_admin_client(admin_client, bootstrap_servers=bootstrap_servers)
        try:
            topics = admin_client.list_topics()
            if topics:
                return {status_key: "Alive"}
        except:
            pass
        return {status_key: "DOWN"}



def dump_pickle(file_path: str, file, mode: str = "wb"):
    import pickle

    with open(file_path, mode=mode) as f:
        pickle.dump(file, f)


def load_pickle(file_path: str, mode: str = "rb", encoding=""):
    import pickle

    with open(file_path, mode=mode) as f:
        return pickle.load(f, encoding=encoding)
